@@ -14,7 +14,7 @@ along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 **********/
 // "liveMedia"
-// Copyright (c) 1996-2011 Live Networks, Inc.  All rights reserved.
+// Copyright (c) 1996-2012 Live Networks, Inc.  All rights reserved.
 // A RTSP server
 // Implementation
 
@@ -156,7 +156,7 @@ int RTSPServer::setUpOurSocket(UsageEnvironment& env, Port& ourPort) {
     // The following statement is enabled by default.
     // Don't disable it (by defining ALLOW_RTSP_SERVER_PORT_REUSE) unless you know what you're doing.
 #ifndef ALLOW_RTSP_SERVER_PORT_REUSE
-    NoReuse dummy; // Don't use this socket if there's already a local server using it
+    NoReuse dummy(env); // Don't use this socket if there's already a local server using it
 #endif
 
     ourSocket = setupStreamSocket(env, ourPort);
@@ -344,7 +344,7 @@ void RTSPServer::RTSPClientSession::reclaimStreamStates() {
 void RTSPServer::RTSPClientSession::resetRequestBuffer() {
   fRequestBytesAlreadySeen = 0;
   fRequestBufferBytesLeft = sizeof fRequestBuffer;
-  fLastCRLF = &fRequestBuffer[-3]; // hack
+  fLastCRLF = &fRequestBuffer[-3]; // hack: Ensures that we don't think we have end-of-msg if the data starts with <CR><LF>
   fBase64RemainderCount = 0;
 }
 
@@ -367,12 +367,13 @@ void RTSPServer::RTSPClientSession::handleAlternativeRequestByte(void* instance,
 
 void RTSPServer::RTSPClientSession::handleAlternativeRequestByte1(u_int8_t requestByte) {
   // Add this character to our buffer; then try to handle the data that we have buffered so far:
-  if (fRequestBufferBytesLeft == 0|| fRequestBytesAlreadySeen >= RTSP_BUFFER_SIZE) return;
+  if (fRequestBufferBytesLeft == 0 || fRequestBytesAlreadySeen >= RTSP_BUFFER_SIZE) return;
   fRequestBuffer[fRequestBytesAlreadySeen] = requestByte;
   handleRequestBytes(1);
 }
 
 void RTSPServer::RTSPClientSession::handleRequestBytes(int newBytesRead) {
+  int numBytesRemaining = 0;
   ++fRecursionCount;
 
   do {
@@ -392,7 +393,8 @@ void RTSPServer::RTSPClientSession::handleRequestBytes(int newBytesRead) {
     unsigned char* ptr = &fRequestBuffer[fRequestBytesAlreadySeen];
 #ifdef DEBUG
     ptr[newBytesRead] = '\0';
-    fprintf(stderr, "RTSPClientSession[%p]::handleRequestBytes() read %d new bytes:%s\n", this, newBytesRead, ptr);
+    fprintf(stderr, "RTSPClientSession[%p]::handleRequestBytes() %s %d new bytes:%s\n",
+	    this, numBytesRemaining > 0 ? "processing" : "read", newBytesRead, ptr);
 #endif
     
     if (fClientOutputSocket != fClientInputSocket) {
@@ -404,7 +406,7 @@ void RTSPServer::RTSPClientSession::handleRequestBytes(int newBytesRead) {
       if (numBytesToDecode > 0) {
 	ptr[newBytesRead] = '\0';
 	unsigned decodedSize;
-	unsigned char* decodedBytes = base64Decode((char*)(ptr-fBase64RemainderCount), decodedSize);
+	unsigned char* decodedBytes = base64Decode((char const*)(ptr-fBase64RemainderCount), decodedSize);
 #ifdef DEBUG
 	fprintf(stderr, "Base64-decoded %d input bytes into %d new bytes:", numBytesToDecode, decodedSize);
 	for (unsigned k = 0; k < decodedSize; ++k) fprintf(stderr, "%c", decodedBytes[k]);
@@ -426,9 +428,8 @@ void RTSPServer::RTSPClientSession::handleRequestBytes(int newBytesRead) {
     }
     
     // Look for the end of the message: <CR><LF><CR><LF>
-    unsigned char *tmpPtr = ptr;
-    if (fRequestBytesAlreadySeen > 0) --tmpPtr;
-    // in case the last read ended with a <CR>
+    unsigned char *tmpPtr = fLastCRLF + 2;
+    if (tmpPtr < fRequestBuffer) tmpPtr = fRequestBuffer;
     while (tmpPtr < &ptr[newBytesRead-1]) {
       if (*tmpPtr == '\r' && *(tmpPtr+1) == '\n') {
 	if (tmpPtr - fLastCRLF == 2) { // This is it:
@@ -452,14 +453,17 @@ void RTSPServer::RTSPClientSession::handleRequestBytes(int newBytesRead) {
     char urlSuffix[RTSP_PARAM_STRING_MAX];
     char cseq[RTSP_PARAM_STRING_MAX];
     unsigned contentLength;
-    if (parseRTSPRequestString((char*)fRequestBuffer, fRequestBytesAlreadySeen,
-			       cmdName, sizeof cmdName,
-			       urlPreSuffix, sizeof urlPreSuffix,
-			       urlSuffix, sizeof urlSuffix,
-			       cseq, sizeof cseq,
-			       contentLength)) {
+    *fLastCRLF = '\0'; // temporarily, for parsing
+    Boolean parseSucceeded = parseRTSPRequestString((char*)fRequestBuffer, fRequestBytesAlreadySeen,
+						    cmdName, sizeof cmdName,
+						    urlPreSuffix, sizeof urlPreSuffix,
+						    urlSuffix, sizeof urlSuffix,
+						    cseq, sizeof cseq,
+						    contentLength);
+    *fLastCRLF = '\r';
+    if (parseSucceeded) {
 #ifdef DEBUG
-      fprintf(stderr, "parseRTSPRequestString() succeeded, returning cmdName \"%s\", urlPreSuffix \"%s\", urlSuffix \"%s\", CSeq \"%s\", Content-Length %u\n", cmdName, urlPreSuffix, urlSuffix, cseq, contentLength);
+      fprintf(stderr, "parseRTSPRequestString() succeeded, returning cmdName \"%s\", urlPreSuffix \"%s\", urlSuffix \"%s\", CSeq \"%s\", Content-Length %u, with %d bytes following the message.\n", cmdName, urlPreSuffix, urlSuffix, cseq, contentLength, ptr + newBytesRead - (tmpPtr + 2));
 #endif
       // If there was a "Content-Length:" header, then make sure we've received all of the data that it specified:
       if (ptr + newBytesRead < tmpPtr + 2 + contentLength) break; // we still need more data; subsequent reads will give it to us 
@@ -486,10 +490,13 @@ void RTSPServer::RTSPClientSession::handleRequestBytes(int newBytesRead) {
       // The request was not (valid) RTSP, but check for a special case: HTTP commands (for setting up RTSP-over-HTTP tunneling):
       char sessionCookie[RTSP_PARAM_STRING_MAX];
       char acceptStr[RTSP_PARAM_STRING_MAX];
-      if (parseHTTPRequestString(cmdName, sizeof cmdName,
-				 urlSuffix, sizeof urlPreSuffix,
-				 sessionCookie, sizeof sessionCookie,
-				 acceptStr, sizeof acceptStr)) {
+      *fLastCRLF = '\0'; // temporarily, for parsing
+      parseSucceeded = parseHTTPRequestString(cmdName, sizeof cmdName,
+					      urlSuffix, sizeof urlPreSuffix,
+					      sessionCookie, sizeof sessionCookie,
+					      acceptStr, sizeof acceptStr);
+      *fLastCRLF = '\r';
+      if (parseSucceeded) {
 #ifdef DEBUG
 	fprintf(stderr, "parseHTTPRequestString() succeeded, returning cmdName \"%s\", urlSuffix \"%s\", sessionCookie \"%s\", acceptStr \"%s\"\n", cmdName, urlSuffix, sessionCookie, acceptStr);
 #endif
@@ -541,8 +548,17 @@ void RTSPServer::RTSPClientSession::handleRequestBytes(int newBytesRead) {
 			      (char const*)fRequestBuffer);
     }
     
+    // Check whether there are extra bytes remaining in the buffer, after the end of the request (a rare case).
+    // If so, move them to the front of our buffer, and keep processing it, because it might be a following, pipelined request.
+    unsigned requestSize = (fLastCRLF+4-fRequestBuffer) + contentLength;
+    numBytesRemaining = fRequestBytesAlreadySeen - requestSize;
     resetRequestBuffer(); // to prepare for any subsequent request
-  } while (0);
+
+    if (numBytesRemaining > 0) {
+      memmove(fRequestBuffer, &fRequestBuffer[requestSize], numBytesRemaining);
+      newBytesRead = numBytesRemaining;
+    }
+  } while (numBytesRemaining > 0);
 
   --fRecursionCount;
   if (!fSessionIsActive) {
@@ -690,6 +706,7 @@ static void parseTransportHeader(char const* buf,
   // First, find "Transport:"
   while (1) {
     if (*buf == '\0') return; // not found
+    if (*buf == '\r' && *(buf+1) == '\n' && *(buf+2) == '\r') return; // end of the headers => not found
     if (_strncasecmp(buf, "Transport: ", 11) == 0) break;
     ++buf;
   }
@@ -697,7 +714,7 @@ static void parseTransportHeader(char const* buf,
   // Then, run through each of the fields, looking for ones we handle:
   char const* fields = buf + 11;
   char* field = strDupSize(fields);
-  while (sscanf(fields, "%[^;]", field) == 1) {
+  while (sscanf(fields, "%[^;\r\n]", field) == 1) {
     if (strcmp(field, "RTP/AVP/TCP") == 0) {
       streamingMode = RTP_TCP;
     } else if (strcmp(field, "RAW/RAW/UDP") == 0 ||
@@ -753,9 +770,10 @@ void RTSPServer::RTSPClientSession
 
   do {
     // First, make sure the specified stream name exists:
+    ServerMediaSession* prevSMS = fOurServerMediaSession;
     fOurServerMediaSession = fOurServer.lookupServerMediaSession(streamName);
     if (fOurServerMediaSession == NULL) {
-      // Check for the special case (noted above), before we up:
+      // Check for the special case (noted above), before we give up:
       if (urlPreSuffix[0] == '\0') {
 	streamName = urlSuffix;
       } else {
@@ -773,7 +791,10 @@ void RTSPServer::RTSPClientSession
       break;
     }
 
-    fOurServerMediaSession->incrementReferenceCount();
+    if (fOurServerMediaSession != prevSMS) {
+      // We're accessing the "ServerMediaSession" for the first time.
+      fOurServerMediaSession->incrementReferenceCount();
+    }
 
     if (fStreamStates == NULL) {
       // This is the first "SETUP" for this session.  Set up our array of states for all of this session's subsessions (tracks):
@@ -1134,10 +1155,11 @@ void RTSPServer::RTSPClientSession
   char* rtpInfo = strDup("RTP-Info: ");
   unsigned i, numRTPInfoItems = 0;
 
-  // Do any required seeking/scaling on each subsession, before starting streaming:
+  // Do any required seeking/scaling on each subsession, before starting streaming.
+  // (However, we don't do this if the "PLAY" request was for just a single subsession of a multiple-subsession stream;
+  //  for such streams, seeking/scaling can be done only with an aggregate "PLAY".)
   for (i = 0; i < fNumStreamStates; ++i) {
-    if (subsession == NULL /* means: aggregated operation */
-	|| subsession == fStreamStates[i].subsession) {
+    if (subsession == NULL /* means: aggregated operation */ || fNumStreamStates == 1) {
       if (sawScaleHeader) {
 	fStreamStates[i].subsession->setStreamScale(fOurSessionId,
 						    fStreamStates[i].streamToken,
@@ -1602,6 +1624,12 @@ UserAuthenticationDatabase::UserAuthenticationDatabase(char const* realm,
 
 UserAuthenticationDatabase::~UserAuthenticationDatabase() {
   delete[] fRealm;
+
+  // Delete the allocated 'password' strings that we stored in the table, and then the table itself:
+  char* password;
+  while ((password = (char*)fTable->RemoveNext()) != NULL) {
+    delete[] password;
+  }
   delete fTable;
 }
 
